// OthelloBoard.java
//
// Informatics 45 Spring 2010
// Code Example: Our Othello Game
//
// This class implements the Othello board, whose job is not only to
// keep track of which cells hold which kinds of tiles, but also to
// encapsulate all of the game logic.  Here, we'll find all of the code
// for determining whether moves are legal, flipping tiles when moves
// are made, counting up the score, and deciding whose turn it is.
//
// The board is a "model" (in the sense of model/view separation that
// we've used often this quarter), so it will notify the view of changes
// using a listener-based mechanism, as we've done several times before.
//
// We made an effort, as we designed this class, to reuse code when we
// could.  In particular, we discovered that deciding whether a move was
// legal required looking in eight different directions for tiles to flip;
// rather than write eight separate loops, we wrote one method that took
// parameters describing the direction we should go.  Similarly, we
// discovered that the code required to actually flip the tiles first
// required us to decide how many tiles to flip, and that this was
// identical to how we decided whether a move was legal; so, we reused
// the code in both places.
//
// I should point out -- and you may recall this from lecture -- that we
// didn't write this code "correctly" at first, in the sense that we didn't
// know all the ways we planned to reuse it.  We adjusted it as we discovered
// new ways to reuse it.  This is an example of something we've talked about
// a lot this quarter: we have to get away from the idea that changing code
// we've already written is a bad thing, or that every line of code gets us
// absolutely closer to our goal.  Writing programs, like writing prose, is
// a discovery process that involves a fair amount of editing along the way.

package inf45.spring2010.examples.othello.game;

import java.util.ArrayList;


public class OthelloBoard
{
	// This constant is used to indicate the width and height, in cells,
	// of the board.  Since we want an 8x8 board, we'll set it to 8.
	private static final int WIDTH_AND_HEIGHT = 8;
	
	// The current state the Othello board is defined with these two
	// fields: a two-dimensional array holding the state of each cell,
	// and a boolean value indicating whether it's the black player's
	// turn (true) or the white player's turn (false).
	private OthelloCell[][] cells;
	private boolean isBlackTurn;

	// Since the board is a model and it needs to notify views of changes,
	// it will employ the standard "listener" mechanism of tracking a list
	// of listeners and firing "events" (by calling methods on those
	// listeners) when things change.
	private ArrayList<OthelloBoardListener> listeners;
	
	
	public OthelloBoard()
	{
		// Arrays, in Java, are objects, which means that variables of
		// array types (like cells, which has type OthelloCell[][]) are
		// really references that say where an array lives.  By default,
		// references point to null.  So we'll need to create an actual
		// two-dimensional array for "cells" to point to.
		cells = new OthelloCell[WIDTH_AND_HEIGHT][WIDTH_AND_HEIGHT];

		// By default, the cells will also have the value null, so we'll
		// need to fill them in with the special value OthelloCell.NONE
		// (part of the OthelloCell enum -- see OthelloCell.java), which
		// we use to indicate an empty cell.
		for (int i = 0; i < WIDTH_AND_HEIGHT; i++)
		{
			for (int j = 0; j < WIDTH_AND_HEIGHT; j++)
			{
				cells[i][j] = OthelloCell.NONE;
			}
		}
				
		// Now we need to put the initial tiles on to the board.  The rules
		// specify that in the four cells around the center of the board,
		// two white and two black tiles (white in upper left and lower right,
		// black in upper right and lower left) should appear.  So we'll put
		// those tiles in place.  Note that we first have to calculate where
		// the center of the board is, since the width and height may vary.
		
		int initialAreaBegin = WIDTH_AND_HEIGHT / 2 - 1;
		
		cells[initialAreaBegin][initialAreaBegin] = OthelloCell.WHITE;
		cells[initialAreaBegin + 1][initialAreaBegin] = OthelloCell.BLACK;
		cells[initialAreaBegin][initialAreaBegin + 1] = OthelloCell.BLACK;
		cells[initialAreaBegin + 1][initialAreaBegin + 1] = OthelloCell.WHITE;

		// Finally, we make a note that it's the black player's turn (because
		// the black player goes first), then create our list of listeners.
		isBlackTurn = true;
		listeners = new ArrayList<OthelloBoardListener>();
	}

	
	// These are fairly standard add/remove listener methods that we've been
	// building all quarter.
	
	public void addOthelloBoardListener(OthelloBoardListener listener)
	{
		listeners.add(listener);
	}
	
	
	public void removeOthelloBoardListener(OthelloBoardListener listener)
	{
		if (listeners.contains(listener))
		{
			listeners.remove(listener);
		}
	}
	
	
	// These are fairly standard "fire event" methods that we've been building
	// all quarter, one corresponding to each method in the listener interface.
	
	private void fireBoardChanged()
	{
		for (OthelloBoardListener listener : listeners)
		{
			listener.boardChanged();
		}
	}
	
	
	private void fireGameOver()
	{
		for (OthelloBoardListener listener : listeners)
		{
			listener.gameOver();
		}
	}
	
	
	// getWidthAndHeight() returns the number of cells wide and high the
	// board is.
	public int getWidthAndHeight()
	{
		return WIDTH_AND_HEIGHT;
	}
	

	// getCell() returns the value in the cell at the given x/y coordinate.
	// Note that the coordinates, here, refer to the coordinates given to
	// the cells themselves (i.e., (0, 0) is the upper-left cell, (4, 3)
	// is the cell four to the right and three down from the upper-left),
	// as opposed to the x/y pixel coordinates where the cells are drawn.
	// The reason why this class deals with cell coordinates and not pixel
	// coordinates is that pixels are part of the view (part of what the
	// board looks like), not part of the model (what the board does, which
	// is what this class implements).
	public OthelloCell getCell(int x, int y)
	{
		if (!isValidCell(x, y))
		{
			throw new IllegalArgumentException(
				"(" + x + ", " + y + ") is not a valid cell");
		}
		
		return cells[x][y];
	}
	

	// isBlackTurn() returns true if it's the black player's turn, and false
	// if it's the white player's turn.
	public boolean isBlackTurn()
	{
		return isBlackTurn;
	}
	

	// getBlackScore() calculates the score for the black player.
	public int getBlackScore()
	{
		return getScore(OthelloCell.BLACK);
	}
	
	
	// getWhiteScore() calculates the score for the white player.
	public int getWhiteScore()
	{
		return getScore(OthelloCell.WHITE);
	}
	

	// getScore() runs through all the cells on the board and counts the
	// number of cells that have a particular value (e.g., BLACK or WHITE).
	// This method uses the naive approach of counting them each time
	// it's called; it might be better to keep track of this score as we
	// go, updating it as tiles are added and flipped.
	private int getScore(OthelloCell cellValue)
	{
		int score = 0;
		
		for (int i = 0; i < WIDTH_AND_HEIGHT; i++)
		{
			for (int j = 0; j < WIDTH_AND_HEIGHT; j++)
			{
				if (cells[i][j] == cellValue)
				{
					score++;
				}
			}
		}
		
		return score;
	}
	

	// makeMove() attempts to make the given move on behalf of the current
	// player.  If the move is not legal, an InvalidMoveException is thrown.
	public void makeMove(int x, int y)
	throws InvalidMoveException
	{
		// If the given move is not valid, throw the exception.
		if (!isValidMove(x, y))
		{
			throw new InvalidMoveException();
		}

		// If we get here, the given move is valid, so we'll flip any
		// tiles need to be flipped in each of the eight directions.
		
		doFlips(x, y, 0, -1);
		doFlips(x, y, 1, -1);
		doFlips(x, y, 1, 0);
		doFlips(x, y, 1, 1);
		doFlips(x, y, 0, 1);
		doFlips(x, y, -1, 1);
		doFlips(x, y, -1, 0);
		doFlips(x, y, -1, -1);

		// After flipping the tiles, we'll add a new tile to the cell where
		// the move was made.
		cells[x][y] =
			isBlackTurn() ? OthelloCell.BLACK : OthelloCell.WHITE;

		
		// Now we'll decide whose turn it is and whether the game is over.
		
		boolean gameOver = false;

		// First, assume it's the other player's turn, since play generally
		// alternates between players.
		isBlackTurn = !isBlackTurn;

		// If the other player has no valid move, we'll switch the turn back
		// to the current player; if the current player also has no move,
		// the game is over.
		if (!hasValidMove())
		{
			isBlackTurn = !isBlackTurn;
			
			if (!hasValidMove())
			{
				gameOver = true;
			}
		}
		
		
		// Finally, we fire events back to the view.  The board has changed,
		// so we fire that first.  Then, if the game is over, we also fire
		// the game over event.

		fireBoardChanged();
		
		if (gameOver)
		{
			fireGameOver();
		}
	}
	
	
	// hasValidMove() returns true if there is a valid move on the board and
	// false if not.  There is a valid move on the board if at least one
	// cell would be a valid move for the current player.
	private boolean hasValidMove()
	{
		for (int x = 0; x < WIDTH_AND_HEIGHT; x++)
		{
			for (int y = 0; y < WIDTH_AND_HEIGHT; y++)
			{
				if (isValidMove(x, y))
				{
					return true;
				}
			}
		}
		
		return false;
	}
	
	
	// isValidMove() returns true if the given move is valid (i.e., it will
	// result in cells being flipped in at least one of the eight directions).
	// Notice that it does its work, primarily, by counting how many flips
	// will be done in each direction.
	private boolean isValidMove(int x, int y)
	{
		// It's not a valid move if there is already a tile in that cell.
		// We forgot this, at first, in lecture, which caused an interesting
		// bug where the game never ended, because we could always just
		// replace an existing tile with a new one.
		if (cells[x][y] != OthelloCell.NONE)
		{
			return false;
		}
		
		return countFlips(x, y, 0, -1) > 0
			|| countFlips(x, y, 1, -1) > 0
			|| countFlips(x, y, 1, 0) > 0
			|| countFlips(x, y, 1, 1) > 0
			|| countFlips(x, y, 0, 1) > 0
			|| countFlips(x, y, -1, 1) > 0
			|| countFlips(x, y, -1, 0) > 0
			|| countFlips(x, y, -1, -1) > 0;
	}
	
	
	// doFlips() flips tiles in a particular direction from the cell
	// whose coordinate is (xStart, yStart).
	private void doFlips(
		int xStart, int yStart, int xDiff, int yDiff)
	{
		// We start by counting how many flips need to be done; we can
		// do this by calling countFlips(), which we also used elsewhere.
		int flips = countFlips(xStart, yStart, xDiff, yDiff);
		
		// Now we'll take the appropriate number of steps in the appropriate
		// direction, flipping tiles as we go.
		int x = xStart;
		int y = yStart;
		
		for (int i = 1; i <= flips; i++)
		{
			x += xDiff;
			y += yDiff;
			flip(x, y);
		}
	}
	

	// This method flips the tile at the given x/y coordinate.  It operates
	// under the assumption that the cell at the given x/y coordinate has
	// either a white or a black tile in it.
	private void flip(int x, int y)
	{
		cells[x][y] =
			getCell(x, y) == OthelloCell.BLACK
				? OthelloCell.WHITE
				: OthelloCell.BLACK;
	}
	
	
	// This method counts the number of tiles that will need to be flipped
	// in a given direction if the current player places a tile at
	// the cell with coordinate (xStart, yStart).  The direction is noted
	// by xDiff and yDiff, which indicate the difference between the
	// coordinates of subsequent cells in the x and y direction; for
	// example, xDiff = 1 and yDiff = -1 means diagonally up and to the
	// right, since y coordinates get smaller as we go up and x coordinates
	// get larger as we go right.
	//
	// We called this method checkFlips() in lecture, which was a fine name
	// when we originally wrote it (because it returned a boolean indicating
	// whether any flips were necessary).  Once we changed its behavior to
	// count the number of tiles needed to be flipped, we probably should have
	// changed its name, too.
	private int countFlips(
		int xStart, int yStart, int xDiff, int yDiff)
	{
		// First, we'll establish who the current player and the opponent is.
		
		OthelloCell myColor =
			isBlackTurn() ? OthelloCell.BLACK : OthelloCell.WHITE;
		
		OthelloCell otherColor =
			isBlackTurn() ? OthelloCell.WHITE : OthelloCell.BLACK;

		// Now we'll start counting the cells to be flipped.  We'll start
		// one cell away from the (xStart, yStart) cell in whichever
		// direction we should be moving; count will indicate how many
		// tiles should be flipped.
		int x = xStart + xDiff;
		int y = yStart + yDiff;
		int count = 0;

		// We'll keep going until we find a cell containing something
		// other than our opponent's color or until we fall off the
		// edge of the board.
		while (isValidCell(x, y) && getCell(x, y) == otherColor)
		{
			x += xDiff;
			y += yDiff;
			count++;
		}

		// If we ended up on a valid cell, found any tiles to flip,
		// and ended up on a cell containing a tile of the current
		// player's color, there are tiles to flip; we'll return
		// the number we counted.  In any other case, we'll return 0.
		if (isValidCell(x, y) && count > 0 && getCell(x, y) == myColor)
		{
			return count;
		}
		else
		{
			return 0;
		}
	}
	
	
	// isValidCell() tells us whether the given x/y coordinate refers to
	// a valid cell.  If the board is, say, 8-by-8, a cell will be valid
	// if x and y are both between 0 and 7 (inclusive).
	private boolean isValidCell(int x, int y)
	{
		return x >= 0 && x < WIDTH_AND_HEIGHT
			&& y >= 0 && y < WIDTH_AND_HEIGHT;
	}
}
