// OthelloBoardPanel.java
//
// Informatics 45 Spring 2010
// Code Example: Our Othello Game
//
// This class implements our custom GUI component that displays the Othello
// board and allows the user to click on its cells in order to make a move.
// We implemented this component by extending JPanel -- a built-in component
// that has no look-and-feel of its own -- and teaching it how to "paint"
// itself.  The "paint" method is called every time a component needs to be
// redrawn; by overriding the paint method with our own implementation, we
// can supply our own code that does the redrawing.

package inf45.spring2010.examples.othello.ui;

import inf45.spring2010.examples.othello.game.InvalidMoveException;
import inf45.spring2010.examples.othello.game.OthelloBoard;
import inf45.spring2010.examples.othello.game.OthelloCell;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


public class OthelloBoardPanel extends JPanel
{
	// We'll draw the board with 20 pixels of empty space around it.  This
	// will make our green-background panel extend beyond the boundaries of
	// the board, which is a nice look.
	private static final int INSET = 20;
	
	// We'll leave a little extra space around the ellipses drawn in each
	// cell.  This value specifies how much.
	private static final int CELL_INSET = 5;

	// In order to draw the board, we'll have to be able to find out some
	// things about the board's state, so we'll store it in this field.
	private OthelloBoard board;
	
	
	public OthelloBoardPanel(final OthelloBoard board)
	{
		this.board = board;
		
		// The range of possible colors that can be used on a display
		// is called a "color space."  There are different color spaces
		// used in computing (and other industries where color is an
		// issue, such as printing), though a common one is called RGB,
		// which stands for "Red-Green-Blue."  The RGB color space is
		// defined by three bytes: a "red" byte, a "green" byte, and a
		// "blue" byte, each of which indicates an amount of red, green,
		// or blue, respectively (with 0 being none and 255 being the
		// maximum amount).  In effect, you're mixing yourself a color
		// from red, green, and blue components.  So, for example, to
		// make a purple color, you can mix red and blue.
		//
		// In our case, we want a nice dark green, so we'll use half
		// of the green available to us.  (If we wanted it darker, we
		// could reduce the green; if we wanted it brighter, we could
		// increase it.)
		//
		// There are other color spaces that are supported by Java (such
		// as RGBA, which includes a fourth component called "alpha" that
		// indicates a level of transparency), but RGB will suit us fine
		// here.
		setBackground(new Color(0, 128, 0));


		// Here, we're adding a mouse listener that will react when someone
		// clicks on our custom panel.  We'll react by calling a method who
		// will decide which cell was clicked on and, if it's a valid one,
		// will attempt to make the move.
		addMouseListener(
			new MouseAdapter()
			{
				public void mouseClicked(MouseEvent e)
				{
					// We'll pass the x/y coordinate (in terms of pixels,
					// relative to the upper-left corner of the panel) where
					// the click occurred.

					handleMouseClick(e.getX(), e.getY());
				}
			});
	}
	
	
	// This is where we describe what the board looks like.  We write a
	// paint() method, with a particular signature (taking a Graphics as
	// a parameter).  This overrides the existing paint() method, allowing
	// us to substitute our own code that will be used to draw the board
	// whenever it needs to be drawn.
	public void paint(Graphics g)
	{
		// This is the first thing an overridden paint() method has to do.
		// I'm a little hazy on the details of why this is -- I think I knew
		// at one point, but no longer remember.
		super.paint(g);

		// Because we're painting on a two-dimensional surface, we can cast
		// our Graphics object to a Graphics2D, allowing us to do more with
		// it.
		Graphics2D g2d = (Graphics2D) g;

		// This is how we turn on antialiasing.  Antialiasing is one of a
		// few different "rendering hints" we can give to Java, which are
		// ways of saying "When you have decisions to make, here are some
		// things I want you to think about."  In this case, we're saying
		// "When you have to decide how to turn something diagonal or round
		// into pixels, use antialiasing to smooth them out."
		g2d.setRenderingHint(
			RenderingHints.KEY_ANTIALIASING,
			RenderingHints.VALUE_ANTIALIAS_ON);

		// setColor() changes the "current color."  Everything we draw will
		// be drawn in that color until we change it to something else; this
		// is akin to picking up a black pen and getting ready to draw with it.
		g2d.setColor(Color.BLACK);

		// The board may vary in size, so we'll find out what its current
		// size is.
		Dimension size = new Dimension();
		getSize(size);

		int boxCount = board.getWidthAndHeight();

		// We'll call methods that can tell us how wide and tall each cell
		// need to be.  (We'll write these methods below.)
		int boxWidth = getBoxWidth();
		int boxHeight = getBoxHeight();

		// First, we draw the horizontal lines around the cells.  To draw
		// a line, you call a method drawLine() that takes four parameters:
		// x1, y1, x2, y2.  The method draws a line from coordinate (x1, y1)
		// to coordinate (x2, y2).
		for (int i = 0; i <= boxCount; i++)
		{
			g2d.drawLine(
				INSET,
				INSET + i * boxHeight,
				INSET + boxCount * boxWidth,
				INSET + i * boxHeight);
		}
		
		// Next, we draw the vertical lines around the cells.
		for (int i = 0; i <= boxCount; i++)
		{
			g2d.drawLine(
				INSET + i * boxWidth,
				INSET,
				INSET + i * boxWidth,
				INSET + boxCount * boxHeight);
		}

		// Finally, we draw ellipses (known, in this context, as "filled
		// ovals") inside the cells that presently have a tile in them.
		// Note that we set the color appropriately (white or black)
		// depending on which kind of tile is in the cell.
		//
		// The fillOval() method draws a filled oval.  We give it four
		// parameters: the x coordinate and y coordinate of the upper-
		// left corner of the oval (i.e., the upper-left corner of a
		// box that bounds the oval) and the width and height of the oval
		// (i.e., the width and height of the box that would bound it).
		for (int i = 0; i < boxCount; i++)
		{
			for (int j = 0; j < boxCount; j++)
			{
				OthelloCell cell = board.getCell(i, j);
				
				if (cell == OthelloCell.WHITE)
				{
					g2d.setColor(Color.WHITE);
				}
				else if (cell == OthelloCell.BLACK)
				{
					g2d.setColor(Color.BLACK);
				}
				
				if (cell != OthelloCell.NONE)
				{
					g2d.fillOval(
						INSET + i * boxWidth + CELL_INSET,
						INSET + j * boxHeight + CELL_INSET,
						boxWidth - CELL_INSET * 2,
						boxHeight - CELL_INSET * 2);
				}
			}
		}
	}
	
	
	// This method finds how wide each box should be, using the current size
	// of the board panel and the width/height of the board (in terms of
	// cells) in its formula.
	private int getBoxWidth()
	{
		Dimension size = new Dimension();
		getSize(size);

		int width = size.width - INSET * 2;
		int boxCount = board.getWidthAndHeight();
		int boxWidth = width / boxCount;
		
		return boxWidth;
	}
	
	
	// This method finds how tall each box should be, using the current size
	// of the board panel and the width/height of the board (in terms of
	// cells) in its formula.
	private int getBoxHeight()
	{
		Dimension size = new Dimension();
		getSize(size);

		int height = size.height - INSET * 2;
		int boxCount = board.getWidthAndHeight();
		int boxHeight = height / boxCount;
		
		return boxHeight;
	}
	
	
	// This method is called whenever the user clicks the mouse on the board
	// panel.  The two parameters, x and y, specify the pixel coordinate within
	// the board panel that was clicked -- so, if the user clicks 30 pixels to
	// the right and 50 pixels below the upper-left corner of the board panel,
	// the parameters will be x=30 and y=50.
	private void handleMouseClick(int x, int y)
	{
		// We derived these formulas in lecture by drawing a couple of pictures
		// and working through them.  We subtract the inset (the pixels of
		// empty space surrounding the board) from each coordinate, then divde
		// by the width/height of each cell to decide which cell was clicked
		// on.  It should be pointed out that this might behave a little
		// strangely at the edges of each cell, but will work fine if the user
		// doesn't click at the edges.
		
		int xCell = (x - INSET) / getBoxWidth();
		int yCell = (y - INSET) / getBoxHeight();
	

		// If the user didn't click on an actual cell, we can just give up
		// handling this mouse click; there's nothing to be done.
		
		if (xCell < 0 || xCell >= board.getWidthAndHeight() ||
			yCell < 0 || yCell >= board.getWidthAndHeight())
		{
			return;
		}

		
		// If the user did click on a cell, we'll ask the board to make the
		// move, then handle any exception thrown if the move wasn't valid.
		
		try
		{
			board.makeMove(xCell, yCell);
		}
		catch (InvalidMoveException ee)
		{
			JOptionPane.showMessageDialog(
				null,
				"That is an invalid move.",
				"Invalid Move",
				JOptionPane.INFORMATION_MESSAGE);
		}
	}
}
