// OthelloFrame.java
//
// Informatics 45 Spring 2010
// Code Example: Our Othello Game
//
// This is a fairly standard frame class, using GridBagLayout to place three
// components into its content pane:
//
// (1) A "score label" at the top, which shows the current score of the game.
//
// (2) A "board panel," which shows the current state of the board and also
//     allows the user to click on the board to make a move.  This is a custom
//     component, implemented in OthelloBoardPanel.java.
//
// (3) A "status label" at the bottom, which shows whose turn it is.

package inf45.spring2010.examples.othello.ui;

import inf45.spring2010.examples.othello.game.OthelloBoard;
import inf45.spring2010.examples.othello.game.OthelloBoardListener;
import java.awt.*;
import javax.swing.*;


public class OthelloFrame extends JFrame implements OthelloBoardListener
{
	private OthelloBoard board;

	private JLabel scoreLabel;
	private OthelloBoardPanel boardPanel;
	private JLabel statusLabel;
	
	
	public OthelloFrame()
	{
		// The frame builds its own model.
		board = new OthelloBoard();

		// We want the frame to receive notifications from the board as its
		// state changes.
		board.addOthelloBoardListener(this);

		setTitle("Informatics 45 Spring 2009: Othello Game");
		setSize(600, 700);
		setResizable(true);
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		getContentPane().setBackground(Color.BLACK);

		buildUI();
		refreshUI();
	}
	
	
	private void buildUI()
	{
		GridBagLayout layout = new GridBagLayout();
		getContentPane().setLayout(layout);

		Font labelFont = new Font("SansSerif", Font.BOLD, 18); 
		
		scoreLabel = new JLabel();
		scoreLabel.setForeground(Color.WHITE);
		scoreLabel.setFont(labelFont);
		layout.setConstraints(
			scoreLabel,
			new GridBagConstraints(
				0, 0, 1, 1, 1.0, 0.0,
				GridBagConstraints.CENTER,
				GridBagConstraints.NONE,
				new Insets(10, 10, 10, 10), 0, 0));
		getContentPane().add(scoreLabel);
		
		boardPanel = new OthelloBoardPanel(board);
		layout.setConstraints(
			boardPanel,
			new GridBagConstraints(
				0, 1, 1, 1, 1.0, 1.0,
				GridBagConstraints.CENTER,
				GridBagConstraints.BOTH,
				new Insets(10, 10, 10, 10), 0, 0));
		getContentPane().add(boardPanel);
		
		statusLabel = new JLabel();
		statusLabel.setForeground(Color.WHITE);
		statusLabel.setFont(labelFont);
		layout.setConstraints(
			statusLabel,
			new GridBagConstraints(
				0, 2, 1, 1, 1.0, 0.0,
				GridBagConstraints.CENTER,
				GridBagConstraints.NONE,
				new Insets(10, 10, 10, 10), 0, 0));
		getContentPane().add(statusLabel);
	}
	
	
	private void refreshUI()
	{
		// Refreshing the UI means to change the text in each of the
		// two labels (the score and the status) and also to ask the
		// board to repaint itself.
		
		scoreLabel.setText(
			"Black: " + board.getBlackScore() +
			"          White: " + board.getWhiteScore());
		
		statusLabel.setText(
			(board.isBlackTurn() ? "Black's" : "White's")
			+ " turn");
		
		boardPanel.repaint();
	}
	
	
	// These are the OthelloBoardListener event-handling methods.

	public void boardChanged()
	{
		// When the board changes, we'll refresh the entire UI.  (There
		// are times when this approach is too inefficient, but it will
		// work fine for our relatively simple UI.)

		refreshUI();
	}
	
	
	public void gameOver()
	{
		// When the game is over, we'll pop up a message box showing the final
		// score, then, after the user dismisses the message box, dispose of
		// this window and end the program.
		
		JOptionPane.showMessageDialog(
			this,
			"Game over!\nFinal score: " + scoreLabel.getText(),
			"Game Over",
			JOptionPane.INFORMATION_MESSAGE);
		
		dispose();
	}
}
